#define HAVE_GETOPT_LONG /* XXX: Autodetect. Autoconf. F33r. --pasky */

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "client/client.h"
#include "client/players.h"
#include "client/render.h"
#include "common/debug.h"
#include "common/globgame.h"
#include "common/lists.h"
#include "common/types.h"
#include "common/xmalloc.h"
#include "gameio/colors.h"



/*** Teams */


/* FIXME: We need TEAMINFO messae which will send info about already
 * existing teams at the time of connect. --pasky */
struct team_info teams[MAX_TEAMS];
int teams_count = 0;

void    init_teams(void)
{
	int i;

	for (i = 0; i < MAX_TEAMS; i++) {
		char buf[MAX_NICKLEN];

		snprintf(buf, MAX_NICKLEN, "Team %d", i);
		memset(&teams[i], 0, sizeof(struct team_info));
		teams[i].name = strdup(buf);
	}
}




/*** Players */


/* Sorted by <score,>deaths (except for LMS, reverse there). */
struct player_info players;
struct player_info *self_player = NULL; /* Ourselves. */
int my_id = -1;
int players_count = 0;

/* Two kills in this period == doublekill. */
#define DOUBLEKILL_PERIOD   1*FPS
/* How long does doublekill last. */
#define DOUBLEKILL_DURATION 4*FPS
#define MULTIKILL_DURATION  6*FPS
long    last_kill   = 0; /* Tick number of the last kill. */
long    double_kill = 0; /* Tick counter; if non-zero, show the message. */
long    multi_kill  = 0; /* Tick counter; if non-zero, show the message. */


void
player_move_up(struct player_info *pos, struct player_info *plr)
{
	if (gstyle != GSTYLE_LMS) {
		while (pos != &players && pos->score < plr->score)
			pos = pos->next;
		while (pos != &players && pos->score == plr->score && pos->deaths > plr->deaths)
			pos = pos->next;
	} else {
		while (pos != &players && pos->deaths > plr->deaths)
			pos = pos->next;
		while (pos != &players && pos->deaths == plr->deaths && pos->score < plr->score)
			pos = pos->next;
	}

	del_from_list(plr);
	add_at_pos(pos->prev, plr);
}

void
player_move_down(struct player_info *pos, struct player_info *plr)
{
	if (gstyle != GSTYLE_LMS) {
		while (pos != &players && pos->score > plr->score)
			pos = pos->prev;
		while (pos != &players && pos->score == plr->score && pos->deaths < plr->deaths)
			pos = pos->prev;
	} else {
		while (pos != &players && pos->deaths < plr->deaths)
			pos = pos->prev;
		while (pos != &players && pos->deaths == plr->deaths && pos->score > plr->score)
			pos = pos->prev;
	}

	del_from_list(plr);
	add_at_pos(pos, plr);
}

struct player_info *
add_player(long id, char *nick, int score, int deaths)
{
	struct player_info *plr = CALLOC(1, sizeof(struct player_info) + strlen(nick));

	plr->id = id;
	plr->score = score;
	plr->deaths = deaths;
	strcpy(plr->nick, nick);

	add_to_list(players, plr);
	player_move_up(plr->next, plr);

	players_count++;
	return plr;
}

void
del_player(struct player_info *plr)
{
	players_count--;
	del_from_list(plr);
	free(plr);
}

struct player_info *
find_player(long id)
{
	struct player_info *plr;

	foreach (plr, players)
		if (plr->id == id)
			return plr;
	return NULL;
}

char *
player_nick_by_id(long id)
{
	struct player_info *plr = find_player(id);

	return plr ? plr->nick : "";
}

void
dead_player(struct player_info *plr)
{
	if (plr->id == self_player->id) {
		last_kill = 0;
		double_kill = 0;
		multi_kill = 0;
	}

	if (gstyle == GSTYLE_LMS) {
		if (plr->deaths > lms_lives) {
			DEBUG("ERROR - player %ld/%s out of its lifespan (%d>%d)",
			      plr->id, plr->nick, plr->deaths, lms_lives);
			return;
		}

		if (plr->deaths == lms_lives - 1 && plr == self_player) {
			msg_add("You have the last chance to prove yourself!");

		} else if (plr->deaths == lms_lives) {
			msg_add("%s drops out of the game, wasting all his %d reincarnations.",
				plr->nick, plr->deaths);
		}
	}
}

void
kill_player(struct player_info *victim, struct player_info *killer)
{
	/* Keep in sync with server code! */

	if (victim != killer && (!victim->team_id || victim->team_id != killer->team_id)) {
		killer->score++;
		victim->deaths++;
		player_move_up(killer->next, killer);
		player_move_down(victim->prev, victim);

		teams[killer->team_id].score++;
		teams[victim->team_id].deaths++;

		if (killer == self_player) {
			if (double_kill || multi_kill) {
				double_kill = 0;
				multi_kill = MULTIKILL_DURATION;
			} else if (tick - last_kill < DOUBLEKILL_PERIOD) {
				double_kill = DOUBLEKILL_DURATION;
			}
			last_kill = tick;
		}

	} else {
		if (killer->score >= 0)
			killer->score--;
		victim->deaths++;
		player_move_down(killer->prev, killer);

		if (teams[killer->team_id].score >= 0)
			teams[killer->team_id].score--;
		teams[victim->team_id].deaths++;
	}

	dead_player(victim);
}



/*** Some visual presentation */

void    render_playerlist(void)
{
	int width = 54;
	struct player_info *plr;
	int line = VIEWPNT_HEIGHT - 2 - players_count - (teams_count ? teams_count + 1 : 0) - 1;
	int i;

	if (teams_count)
		width += 9;

	if (line < 0) line = 0;
	if (line <= msgshow) {
		line++;
		msgshow = line - 1;
	} else if (line > msgshow + 1) {
		screen_darken(2, line - 1, width, 8);
	}

	screen_darken(2, line, width, COLOR_DARKGRAY);
	screen_printf(2, line, COLOR_WHITE, "         %s%-3s  |  %-15s  |  %-5s  |  %5s  ",
	              teams_count ? "Team  |  " : "",
	              "ID", "Player", "Frags", gstyle == GSTYLE_LMS ? "Lives" : "Deaths");
	
	line++;

	foreachback (plr, players) {
		int pcolor = plr_color(plr->team_id ? 7 + plr->team_id : plr->id);

		screen_darken(2, line, width, 8);
		if (teams_count)
			screen_printf(2, line, COLOR_WHITE, "%c(ooo)~~%c  %-4d  |  %-3d  |  %-15s  |  %5d  |  %5d  ",
					PRINTF_COLOR(pcolor), PRINTF_COLOR(0),
				      plr->team_id, plr->id, plr->nick, plr->score,
				      gstyle == GSTYLE_LMS ? lms_lives - plr->deaths : plr->deaths);
		else
			screen_printf(2, line, COLOR_WHITE, "%c(ooo)~~%c  %-3d  |  %-15s  |  %5d  |  %5d  ",
					PRINTF_COLOR(plr_color(plr->id)), PRINTF_COLOR(0),
				      plr->id, plr->nick, plr->score,
				      gstyle == GSTYLE_LMS ? lms_lives - plr->deaths : plr->deaths);
		line++;
		if (line > VIEWPNT_HEIGHT - 1)
			return;
	}

	if (teams_count) {
		screen_darken(2, line++, width, 8);

		for (i = 1; i < MAX_TEAMS; i++) {
			if (!teams[i].active) continue;

			screen_darken(2, line, width, 8);
			if (gstyle == GSTYLE_CTF)
				screen_printf(2, line, COLOR_WHITE, "%c(ooo)~~%c       %-4d     |  %-15s  |  %2d/%-2d  |  %2d/%-2d  ",
					      PRINTF_COLOR(plr_color(7+i)), PRINTF_COLOR(0),
					      i, teams[i].name, teams[i].score, teams[i].deaths,
					      teams[i].fwon, teams[i].flost);
			else
				screen_printf(2, line, COLOR_WHITE, "%c(ooo)~~%c       %-4d     |  %-15s  |  %5d  |  %5d  ",
					      PRINTF_COLOR(plr_color(7+i)), PRINTF_COLOR(0),
					      i, teams[i].name, teams[i].score, teams[i].deaths);
			line++;
			if (line > VIEWPNT_HEIGHT - 1)
				return;
		}
	}

	screen_darken(2, line, width, 8);
}
