#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include "common/packet.h"
#include "common/debug.h"
#include "server/server.h"

static struct player *
find_player(u16 id)
{
	struct player *p;

	foreach(p, players)
		if (p->id == id)
			return p;

	return NULL;
}

int	udppck_recv (struct uconn *server, struct udp_packet *pck)
{
	static u8 buffer[PACKET_SIZE];
	u8	*data = buffer;
	int	len;
	u16	mysum, pcksum;
	

	len = udp_recv(server, buffer, PACKET_SIZE-1, &(pck->from));
	if (len < 0) {
		if (errno != EAGAIN)
			fprintf(stderr, "Error receiving UDP packet: %s\n", strerror(errno));
		return 0;
	}
	if (len == 0) {
		DEBUG("empty packet");
		return 0;
	}
	if (len <= 2) {
		DEBUG("packet too small/NULL");
		return 0;
	}
	pcksum = UNPACKW(data); data += 2; len-=2;
	mysum = in_chksum((u16 *)(data), len);

	if (pcksum != mysum) {
		DEBUG("bad checksum, dropping (%04x vs %04x)", pcksum, mysum);
		return 0;
	}

	pck->type = *data++;
	DEBUG("pck: %x", pck->type);
	switch(pck->type) {
		case HC_HELLO: {
				u16 id = UNPACKW(data);
				struct player *p;
				DEBUG("%d", id);
				p = find_player(id);
				if (!p) return 1;
				assert(!player_is_ai(p));
				p->udp = udp_dup(server, pck->from.sin);
			}
			
		case HC_START:
		case HS_HELLO:
		case HS_END:
		case HS_ERROR:
			return 1;
		case HC_UPDATE:
			if (len < 4) {
				DEBUG("packet too small for HC_UPDATE");
				return 0;
			}
			pck->cu.ackseq = UNPACKW(data);
			data += 2;
			pck->cu.direction = *data++;
			return 1;
		case HS_UPDATE:
			if (len < 7) {
				DEBUG("packet too small for HS_UPDATE");
				return 0;
			}
			pck->su.curseq = UNPACKW(data);
			data += 2;
			pck->su.diffseq = UNPACKW(data);
			data += 2;
			pck->su.size = UNPACKW(data);
			data += 2;
			pck->su.data = data;
			if (len - (1+3*2) != pck->su.size) {
				DEBUG("malformed packet: datalen=%d != difflen=%d", len-(1+3*2), pck->su.size);
				return 0;
			}
			//DEBUGDMP(pck->su.data, pck->su.size);
			return 1;
		default:
			DEBUG("unknown packet type %d", pck->type);
			return 0;
	}
}

int	pck_send(struct uconn *c, u8 *data, u32 len)
{
	u16 	sum;
	sum = in_chksum((u16 *)(data+2), len-2);
	PACKW(data, sum);
	return udp_send(c, data, len);
}

int	udppck_send_opcode (struct uconn *c, u8 opcode)
{
	u8	buffer[3];
	buffer[2] = opcode;
	return pck_send(c, buffer, 3);
}

int	udppck_send_cupdate (struct uconn *c, u16 ack, u8 dir)
{
	u8 	buffer[6];
	buffer[2] = HC_UPDATE;
	buffer[3] = ack&0xFF;
	buffer[4] = ack>>8;
	buffer[5] = dir;
	return pck_send(c, buffer, 6);
}

int	udppck_send_supdate (struct uconn *c, u16 curseq, u16 diffseq, u16 size, u8 *diff)
{
	u8	buffer[PACKET_SIZE], *ptr = buffer+2;
	if (size + 7 + 2 > PACKET_SIZE) {
		DEBUG("update too big");
		return 0;
	}
	*ptr++ = HS_UPDATE;
	PACKW(ptr, curseq);
	ptr += 2;
	PACKW(ptr, diffseq);
	ptr += 2;
	PACKW(ptr, size);
	ptr += 2;
	memcpy(ptr, diff, size);
	return pck_send(c, buffer, size+7+2);
}

int	udppck_send_hello(struct uconn *c, u16 id)
{
	char	buf[5];
	buf[2] = HC_HELLO;
	PACKW((&buf[3]), id);
	DEBUG("sending udp hello to establish nat, id=%d\n", id);
	return pck_send(c, buf, 5);
}

/*	stolen from some udp/icmp flooder...
	sorry dear smurfer :)
*/
u16	in_chksum(u16 *ipbuf, u32 iplen )
{
	register u32 nleft = iplen;
	register u32 sum = 0;
	u_short answer = 0;

	while( nleft > 1 ) {
		sum += *ipbuf++;
		nleft -= 2;
	}

	if( nleft == 1 ) 
		sum += *(u8 *)(ipbuf);

    	sum = ( sum >> 16 ) + ( sum + 0xffff );
	sum += ( sum >> 16 );
	answer = ~sum;

	return (answer);
}

