/*
 * linux keyboard driver.
 * -sd
 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#include <linux/vt.h>
#include "gameio/gameio.h"

static	struct	termios oldflags;
static	int	tty_fd;
static	int	oldkbmode;

#define KB_MODIFIERS    (RAWKB_SHIFT_FLAG | RAWKB_CTRL_FLAG | RAWKB_ALT_FLAG | \
			 RAWKB_LWIN_FLAG | RAWKB_RWIN_FLAG | RAWKB_MENU_FLAG)
#define KB_LED_FLAGS    (RAWKB_SCROLOCK_FLAG | RAWKB_NUMLOCK_FLAG | RAWKB_CAPSLOCK_FLAG)

/* lookup table for converting kernel scancodes into our format - stolen from allegro */
static unsigned char kernel_to_mycode[128] =
{
   /* 0x00 */  0,		  RAWKEY_ESC,        RAWKEY_1,          RAWKEY_2,
   /* 0x04 */  RAWKEY_3,          RAWKEY_4,          RAWKEY_5,          RAWKEY_6,
   /* 0x08 */  RAWKEY_7,          RAWKEY_8,          RAWKEY_9,          RAWKEY_0,
   /* 0x0C */  RAWKEY_MINUS,      RAWKEY_EQUALS,     RAWKEY_BACKSPACE,  RAWKEY_TAB,
   /* 0x10 */  RAWKEY_Q,          RAWKEY_W,          RAWKEY_E,          RAWKEY_R,
   /* 0x14 */  RAWKEY_T,          RAWKEY_Y,          RAWKEY_U,          RAWKEY_I,
   /* 0x18 */  RAWKEY_O,          RAWKEY_P,          RAWKEY_OPENBRACE,  RAWKEY_CLOSEBRACE,
   /* 0x1C */  RAWKEY_ENTER,      RAWKEY_LCONTROL,   RAWKEY_A,          RAWKEY_S,
   /* 0x20 */  RAWKEY_D,          RAWKEY_F,          RAWKEY_G,          RAWKEY_H,
   /* 0x24 */  RAWKEY_J,          RAWKEY_K,          RAWKEY_L,          RAWKEY_COLON,
   /* 0x28 */  RAWKEY_QUOTE,      RAWKEY_TILDE,      RAWKEY_LSHIFT,     RAWKEY_BACKSLASH,
   /* 0x2C */  RAWKEY_Z,          RAWKEY_X,          RAWKEY_C,          RAWKEY_V,
   /* 0x30 */  RAWKEY_B,          RAWKEY_N,          RAWKEY_M,          RAWKEY_COMMA,
   /* 0x34 */  RAWKEY_STOP,       RAWKEY_SLASH,      RAWKEY_RSHIFT,     RAWKEY_ASTERISK,
   /* 0x38 */  RAWKEY_ALT,        RAWKEY_SPACE,      RAWKEY_CAPSLOCK,   RAWKEY_F1,
   /* 0x3C */  RAWKEY_F2,         RAWKEY_F3,         RAWKEY_F4,         RAWKEY_F5,
   /* 0x40 */  RAWKEY_F6,         RAWKEY_F7,         RAWKEY_F8,         RAWKEY_F9,
   /* 0x44 */  RAWKEY_F10,        RAWKEY_NUMLOCK,    RAWKEY_SCRLOCK,    RAWKEY_7_PAD,
   /* 0x48 */  RAWKEY_8_PAD,      RAWKEY_9_PAD,      RAWKEY_MINUS_PAD,  RAWKEY_4_PAD,
   /* 0x4C */  RAWKEY_5_PAD,      RAWKEY_6_PAD,      RAWKEY_PLUS_PAD,   RAWKEY_1_PAD,
   /* 0x50 */  RAWKEY_2_PAD,      RAWKEY_3_PAD,      RAWKEY_0_PAD,      RAWKEY_DEL_PAD,
   /* 0x54 */  0,		  0,		     RAWKEY_BACKSLASH2, RAWKEY_F11,
   /* 0x58 */  RAWKEY_F12,        0,	             0,		        0,
   /* 0x5C */  0,		  0,		     0,			0,
   /* 0x60 */  RAWKEY_ENTER_PAD,  RAWKEY_RCONTROL,   RAWKEY_SLASH_PAD,  RAWKEY_PRTSCR,
   /* 0x64 */  RAWKEY_ALTGR,      0,		     RAWKEY_HOME,       RAWKEY_UP,
   /* 0x68 */  RAWKEY_PGUP,       RAWKEY_LEFT,       RAWKEY_RIGHT,      RAWKEY_END,
   /* 0x6C */  RAWKEY_DOWN,       RAWKEY_PGDN,       RAWKEY_INSERT,     RAWKEY_DEL,
   /* 0x70 */  0,		  0,		     0,	                0,
   /* 0x74 */  0,		  0,		     0,		        RAWKEY_PAUSE,
   /* 0x78 */  0,		  0,		     0,		        0,
   /* 0x7C */  0,		  RAWKEY_LWIN,       RAWKEY_RWIN,       RAWKEY_MENU
};

static unsigned short modifier_table[RAWKEY_MAX - RAWKEY_MODIFIERS] =
{
   RAWKB_SHIFT_FLAG,    RAWKB_SHIFT_FLAG,    RAWKB_CTRL_FLAG,
   RAWKB_CTRL_FLAG,     RAWKB_ALT_FLAG,      RAWKB_ALT_FLAG,
   RAWKB_LWIN_FLAG,     RAWKB_RWIN_FLAG,     RAWKB_MENU_FLAG,
   RAWKB_SCROLOCK_FLAG, RAWKB_NUMLOCK_FLAG,  RAWKB_CAPSLOCK_FLAG
};

#define NUM_PAD_KEYS 17
static char pad_asciis_numlock[NUM_PAD_KEYS] = "0123456789+-*/\r,.";
static char pad_asciis_no_numlock[NUM_PAD_KEYS] = { 
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
	'+', '-', '*', '/', '\r', -1, -1
};

static void kbd_input(unsigned char *buf, int len) 
{
	int	i;

	for (i = 0; i < len; i++, buf++) {
		int press, kcode, code, map, ascii, val;
		struct kbentry kbe;

		press = !(*buf & 0x80);
		kcode = *buf & 0x7f;
		code = kernel_to_mycode[kcode];


                if (code >= RAWKEY_MODIFIERS) {
                        int flag = modifier_table[code - RAWKEY_MODIFIERS];
                        if (press) {
                                if (flag & KB_MODIFIERS)
                                        rawmods |= flag;
                                else if ((flag & KB_LED_FLAGS))
                                        rawmods ^= flag;
                        } else {
                                if (flag & KB_MODIFIERS)
                                        rawmods &= ~flag;
                        }
			gio_handle_raw(code, -1, press);
                        continue;
                }


                /* Build kernel keymap number */
                map = 0;
                if (rawkeys[RAWKEY_LSHIFT] || rawkeys[RAWKEY_RSHIFT]) map |= 1;
                if (rawkeys[RAWKEY_ALTGR]) map |= 2;
                if (rawkeys[RAWKEY_LCONTROL] || rawkeys[RAWKEY_RCONTROL]) map |= 4;
                if (rawkeys[RAWKEY_ALT]) map |= 8;

		/* translate to kernel's kb value */
                kbe.kb_table = map;
                kbe.kb_index = kcode;
                ioctl(tty_fd, KDGKBENT, &kbe);
		
		switch (code) {
			case RAWKEY_BACKSPACE:
				ascii = 8;
				break;
			default:
				ascii = KVAL(kbe.kb_value);
        			if (kbe.kb_value == K_NOSUCHMAP)
					ascii = 0;
			
		}

		switch (KTYP(kbe.kb_value)) {
			case KT_CONS:
				if (!ioctl (tty_fd, VT_ACTIVATE, KVAL(kbe.kb_value) + 1)) {
					gio_handle_raw (code, -1, press);
					continue;
				}
				ascii = 0;
				break;
			case KT_LETTER:
				/* Apply capslock translation */
				if (rawmods & RAWKB_CAPSLOCK_FLAG)
					ascii ^= 0x20;
                                break;
                        case KT_LATIN:
                        case KT_ASCII:
                                break;
			case KT_PAD:
				ascii = -1;
				val = KVAL(kbe.kb_value);
				if ((val >= 0) && (val < NUM_PAD_KEYS)) {
					if (rawmods & RAWKB_NUMLOCK_FLAG)
						ascii = pad_asciis_numlock[val];
					else
						ascii = pad_asciis_no_numlock[val];
				}
                                break;
                        case KT_SPEC:
				if (code == RAWKEY_ENTER) {
                                        ascii = '\r';
                                        break;
                                }
                        case KT_FN:
                        case KT_CUR:
                                ascii = -1;
                                break;
		}
		gio_handle_raw (code, ascii, press);		
	}
}

/********************************** PUBLIC ******************************/
static	int	getkey()
{
	char	buf[128];
	int	n;
	struct	timeval tv;
	
	fd_set rfds;
	FD_ZERO(&rfds);
	FD_SET(tty_fd, &rfds);
	tv.tv_sec = tv.tv_usec = 0;
	if (select(tty_fd+1, &rfds, NULL, NULL, &tv) != 1)
		return -1;
	n = read(tty_fd, buf, 128);
	if (n > 0)
		kbd_input(buf, n);
	return -1; /* fail if unintentionally called as non-raw kbd */
}

static	int	cleanup()
{
	ioctl(tty_fd, KDSKBMODE, oldkbmode);
	tcsetattr(tty_fd, TCSANOW, &oldflags);
	return 0;
}

static	int	init(int fd)
{
	static	struct	termios newflags;
	
	if (ioctl(fd, KDGKBMODE, &oldkbmode))
		GIO_ERROR("failed to detect initial keyboard mode");

	tcgetattr(fd, &oldflags);
	newflags = oldflags;
	newflags.c_lflag = 0;
	newflags.c_iflag = 0;
	newflags.c_cflag = CS8;
	if (tcsetattr(fd, TCSANOW, &newflags))
		GIO_ERROR("failed to setup terminal flags");


	if (ioctl(fd, KDSKBMODE, K_MEDIUMRAW) < 0) {
		tcsetattr(fd, TCSANOW, &oldflags);
		GIO_ERROR("failed switching the keyboard into raw mode");
	}
	tty_fd = fd;
	return 0;
}

/* driver descriptor */
struct	gi_kbd_driver linux_kbd = {
	.name	=	"linux",
	.desc	=	"raw Linux virtual terminal keyboard",
	.raw	=	1,
	.init	=	init,
	.cleanup=	cleanup,
	.getkey	=	getkey
};
