#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#include "common/debug.h"
#include "common/math.h"
#include "common/object.h"
#include "common/surface.h"
#include "common/xmalloc.h"
#include "server/console.h"
#include "server/game.h"
#include "server/players.h"
#include "server/server.h"


/* Some struct ai, rather simple for now. */
/* See .h for description of the states.
 *
 * TODO:
 * * Getting around larger stone blocks (i.e. home walls).
 * * Difficulty levels - adjusts speed, intelligence, aggressivity
 * * Characters, for that matter - each bot has those attributes laid
 *                                 differently (sum of them being const, though)
 * * Some strategy? Defensive - hiding behind stone and in homes,
 *                              even tricks to move the enemy's attention
 *                              to another bot?
 *                  Agressive - ambushing in front of home, getting the start
 *                              of an enemy?
 *                  Cooperative - short time defensive/agressive alliances
 *                                (based on score?)
 * * Communication - have an intelligent conversation about weapons,
 *                   mobile warfare, culture and quantum physics
 *                   with the player.
 */


/* Operational parameters. Mostly guessing. */

/* Maximal distance on which it is worth shooting. */
#define GOOD_SHOT_RADIUS    600
/* How far to back off. 24 is =~ sqrt(16*16+16*16), 16 is height of one
 * character. */
#define BACKOFF_DISTANCE    24
/* How further must be the pursuee than the nearest potential victim before
 * we retarget. */
#define RETARGET_RATIO      3
/* Once per how many ticks are we allowed to change direction when hunting? */
#define HUNT_TURN_PERIOD    FPS/2
/* Once per how many ticks are we allowed to change direction when backing off? */
#define BACKOFF_TURN_PERIOD FPS/4

/* Threshold values before switching to AI_FLEE. */
#define MIN_NORMAL_ENERGY   MAX_ENERGY/3
#define MIN_NORMAL_SHIELDS  MAX_SHIELDS/3
/* How long after being shot last time should the "revenge" last. */
#define MAX_REVENGE_TICKS   FPS /* 1s */
/* Threshold values in revenge mode. */
#define MIN_REVENGE_ENERGY  MAX_ENERGY/4
#define MIN_REVENGE_SHIELDS MAX_SHIELDS/4
/* Threshold values before switching to critical mode (slow speed). */
#define MAX_CRITICAL_ENERGY MAX_ENERGY/6
#define MAX_CRITICAL_SHIELDS 0

/* Angle fuzziness - probabily of changing angle by 1 or 2 respectively. */
#define STUCK_FUZZINESS1    7
#define STUCK_FUZZINESS2    9
#define HUNT_FUZZINESS1     FPS /* ... if ever? */
#define HUNT_FUZZINESS2     FPS*2


/* Local utility functions */

#if 0
#define AIDEBUG(x...) DEBUG("[ai] " x);
#else
#define AIDEBUG(x...)
#endif

#define in_revenge(bot_) (bot_->ai.last_shot <= MAX_REVENGE_TICKS)

#define fine_surface(x_,y_) (surface[x_ / XPOINTS_PERCHAR][y_ / YPOINTS_PERCHAR])

/* Determines whether <from> and <target> are well-aligned for shooting. */
static int
good_shot (struct tank *from, struct tank *target)
{
	int    fromx, fromy; /* Tank coords */
	struct object bullet;  /* Our virtual bullet */

	bullet.gfxid = GFX_BULLET_ID;

	bullet.angle = from->obj.angle;
	get_vector(bullet.angle, BULLET_SPEED, &bullet);

	sprite_border(from->obj.spos.sprite, from->obj.angle, &fromx, &fromy);
	bullet.x = fromx = from->obj.x + fromx*XPOINTS_PERCHAR;
	bullet.y = fromy = from->obj.y + fromy*YPOINTS_PERCHAR;
	obj_sync_sprite(&bullet);

	while (!tank_in_region(target, bullet.spos.sx, bullet.spos.sy,
                               bullet.spos.sprite->width, bullet.spos.sprite->height)) {
		bullet.x += bullet.vx;
		bullet.y += bullet.vy;
		obj_sync_sprite(&bullet);

		/* We allow shooting into dirt - we will eventually
		 * dig to the enemy that way too and it speeds up
		 * our approach. We don't mind shooting "through"
		 * other players too, which needs a FIXME when (if)
		 * we implement botpacks. */
		if (distance(fromx, fromy, bullet.x, bullet.y) > GOOD_SHOT_RADIUS * GOOD_SHOT_RADIUS
		    || surface_calc_collision(&bullet.spos, G_STONE))
			return 0;
	}

	return 1;
}

/* Scan surface and find the nearest tank there. Returns the player. */
static struct player *
find_nearest_enemy(struct player *bot)
{
	int sx = bot->tank->obj.x/XPOINTS_PERCHAR;
	int sy = bot->tank->obj.y/YPOINTS_PERCHAR;
	int minx = sx - VIEWPNT_WIDTH/2;
	int miny = sy - VIEWPNT_HEIGHT/2;
	int maxx = sx + VIEWPNT_WIDTH/2;
	int maxy = sy + VIEWPNT_HEIGHT/2;
	struct player *best = NULL;
	int bestd = INT_MAX;
	int x, y;

	for (x = minx; x < maxx; x++)
		for (y = miny; y < maxy; y++) {
			struct player *p;
			int d;

			if (!(surface[x][y] & G_PLAYER))
				continue;

			d = distance(sx, sy, x, y);
			if (d >= bestd)
				continue;

			p = find_player_by_xy(x, y, 1, 1);
			/* FIXME: Sometimes, p is NULL here. Why?! --pasky */
			if (!p || p == bot || !tank_ready(p->tank)
			    || (bot->team && p->team == bot->team))
				continue;

			best = p; bestd = d;
		}

	return best;
}

/* Determines the proper angle of <from> heading to <target>. */
static int
get_angle (int from_x, int from_y, int target_x, int target_y)
{
	/* XXX: Do we get it right even for short distances? */

	/* Yes, I could probably construct a single huge logical expression
	 * for this. No, I couldn't understand it in 20 seconds then. */

	/* Assuming tanks are always at least 3 chars wide: */
	if (abs(from_y - target_y) <= 1*YPOINTS_PERCHAR)
		return 2 + (from_x < target_x ? 0 : 4);
	if (abs(from_x - target_x) <= 1*XPOINTS_PERCHAR)
		return 0 + (from_y < target_y ? 4 : 0);

	if (from_x > target_x) {
		return 4 + (from_y < target_y ? 1 : 3);
	} else {
		return 0 + (from_y < target_y ? 3 : 1);
	}
}

/* Add some randomization to the angle. */
static int
fuzzy_angle (int angle, int f1, int f2)
{
	return (angle + 8 /* compensate for possible negatives */
	        + (mkrand(f1) == 0 ? mkrand(2) - 1 : 0)
	        + (mkrand(f2) == 0 ? mkrand(4) - 2 : 0)) % 8;
}

static void
yell(struct player *src, struct player *target, int perc,
     const char **quips, int quipscount)
{
	static char buf[256];

	if (mkrand(100) >= perc)
		return;
	snprintf(buf, 100, quips[mkrand(quipscount)], target ? target->name : NULL);
	bcast_say(src, buf, strlen(buf));
}


/* External struct ai control */

void    reset_ai (struct ai *ai)
{
	ai->state = AI_IDLE;
	ai->idle.tick_count = 2 * FPS;
	ai->backing_off = 0;
	ai->last_shot = 100000;
	ai->last_turnaround = 100000;
}

void    stop_ai_pursue (struct player *p)
{
	struct player *bot;

	/* Let the pursuers mind their own bussiness. */
	foreach (bot, players)
		if (bot->ai.state == AI_HUNT && bot->ai.hunt.pursuing == p) {
			bot->ai.state = AI_EXPLORE;
			bot->ai.explore.tick_count = 0;
		}
}


/* struct ai hooks */

void    hit_ai_hook (struct player *victim, struct player *trespasser)
{
	static const char *quips[] = {
		"%s: Yer dead, sucker!", "Don'na ya dare, %s!", "Naw ya'l see, %s!",
		"%s: Watcha for yer death, sux0r!", "%s: Last mistake in yer life, pal.",
	};

	if (!tank_ready(victim->tank))
		return;

	/* TODO: Keep some relation to all players and change the pursuee
	 * only if relation difference is over certain threshold (increase
	 * relation for each hostile act, decrease (parabolically) over time).
	 * --pasky */
	if (victim->tank->energy >= MIN_REVENGE_ENERGY &&
	    victim->tank->shields >= MIN_REVENGE_SHIELDS) {
		if (victim->ai.state != AI_HUNT || victim->ai.hunt.pursuing != trespasser)
			yell(victim, trespasser, 5, quips, 5);
		victim->ai.state = AI_HUNT;
		victim->ai.hunt.pursuing = trespasser;
	}
	victim->ai.last_shot = 0;
}

void    stuck_ai_hook (struct player *victim)
{
	if (!tank_ready(victim->tank))
		return;

	/* Don't back off if we can shoot anyway. We are probably
	 * head-to-head with the enemy. */
	if (victim->ai.state == AI_HUNT
	    && good_shot(victim->tank, victim->ai.hunt.pursuing->tank))
		return;

	if (victim->ai.last_turnaround <= BACKOFF_TURN_PERIOD)
		return;
	victim->ai.last_turnaround = 0;

	/* Don't back off if exploring - let us just go in another direction. */
	if (victim->ai.state == AI_EXPLORE) {
		victim->ai.explore.tick_count = 0;
		return;
	}

	if (!victim->ai.backing_off) {
		AIDEBUG("[%d] gonna back off...", victim->id);
		victim->ai.backoff_x = victim->tank->obj.x;
		victim->ai.backoff_y = victim->tank->obj.y;
	}
	victim->ai.backing_off++;

	AIDEBUG("[%d] backoff old angle %d", victim->id,
		victim->tank->next_angle);
	/* Try some other direction (usually _almost_ opposite); we always
	 * use next_angle instead of angle for maximal variability
	 * of directions if we are consistently getting stuck in
	 * one place. */
	victim->tank->next_angle = fuzzy_angle((mkrand(2) ? 3 : 5)
	                                      + victim->tank->next_angle,
					      STUCK_FUZZINESS1, STUCK_FUZZINESS2);
	AIDEBUG("[%d] backoff new angle %d", victim->id,
		victim->tank->next_angle);
}

void    kill_ai_hook (struct player *killer, struct player *victim)
{
	static const char *quips[] = {
		"Hah, sux0r!", "Me told ya, %s. >:)", "N4w4 y4mm4 p4w!",
		"HAR HAR!", "LmAo!", "Think again next time, pal.", "Excellent!"
	};

	if (!tank_ready(victim->tank))
		return;

	yell(killer, victim, 20, quips, 7);
	killer->ai.state = AI_EXPLORE;
	killer->ai.explore.tick_count = 0;
}

void    death_ai_hook (struct player *victim, struct player *killer)
{
	static const char *quips[] = {
		"Fux0rz y33, %s!", "Fuck you, %s! >:(", "Darn!",
		"That was tight!", "Whoops.", "You'll have less luck next time, pal.",
		"Watcha for my rew3nge, little sucker!",
	};

	if (!killer || killer == victim || !tank_ready(victim->tank))
		return;
	yell(victim, killer, 20, quips, 7);
}


static void flee_home(struct player *bot)
{
	int x, y1, y2;
	static const char *quips[] = {
		"Clean way, clean way! Emergency!", "Shit!", "Oops...",
		"This doesn't look so good, actually. :/",
		/* Hunt mode only: */
		"M'kay, ya can have this one, %s.", "%s: Damned luck you have, pal.",
		"Eek, I think I'll better go.",
	};

	yell(bot, bot->ai.state == AI_HUNT ? bot->ai.hunt.pursuing : NULL,
	     5 + (bot->ai.state == AI_HUNT ? 10 : 0),
	     quips, 4 + (bot->ai.state == AI_HUNT ? 3 : 0));

	bot->ai.backing_off = 0;
	bot->ai.state = AI_FLEE;

	/* TODO: Find the nearest home if it's only matter of energy. */

	/* We decide whether it's nearer to the south entry or the north
	 * entry. */
	/* TODO: If home wall in the way, set the target to the home
	 * corner first. --pasky */
	bot->ai.flee.home_x = x = bot->home->obj.x;
	y1 = bot->home->obj.y - gfx_home->directions[0]->frames[0]->center_y*YPOINTS_PERCHAR;
	y2 = bot->home->obj.y + gfx_home->directions[0]->frames[0]->center_y*YPOINTS_PERCHAR;
	if (distance(bot->tank->obj.x, bot->tank->obj.y, x, y1)
	    < distance(bot->tank->obj.x, bot->tank->obj.y, x, y2))
		bot->ai.flee.home_y = y1;
	else
		bot->ai.flee.home_y = y2;
}


/* State update_bot handlers */

typedef void (*state_hook)(struct player *bot);
void bot_idle(struct player *bot);
void bot_explore(struct player *bot);
void bot_hunt(struct player *bot);
void bot_flee(struct player *bot);

void    bot_idle (struct player *bot)
{
	AIDEBUG("[%d] idle.", bot->id);
	/* *Yawn*. Make sure we aren't madly running around or something. */
	bot->tank->speed = 0;
	bot->tank->fire = 0;

	if (--bot->ai.idle.tick_count < 0
	    && (!surface_calc_collision(&bot->tank->obj.spos, G_HOME)
	        || (bot->tank->energy >= MAX_ENERGY
	            && bot->tank->shields >= MAX_SHIELDS))) {
		bot->ai.state = AI_EXPLORE;
		bot->ai.explore.tick_count = 0;
	}
}

void    bot_explore (struct player *bot)
{
	struct player *enemy;
	int speed = TANK_SPEED;

	AIDEBUG("[%d] Exploring [%d,%d]->[%d]...", bot->id,
	        bot->tank->obj.x, bot->tank->obj.y, bot->tank->obj.angle);

	bot->ai.explore.tick_count--;

	/* We are in the middle of our home - get us out! */
	if (bot->ai.explore.tick_count < 0
	    && abs(bot->tank->obj.x - bot->home->obj.x) <= 2*XPOINTS_PERCHAR
	    && abs(bot->tank->obj.y - bot->home->obj.y) <= 1*YPOINTS_PERCHAR) {
		bot->tank->next_angle = 4 * mkrand(2);
		bot->ai.explore.tick_count = FPS / 2; /* ?more intelligence! */
		return;
	}

	/* Look for an enemy, if not inside home. */
	if (!(fine_surface(bot->tank->obj.x, bot->tank->obj.y) & G_HOME)) {
		enemy = find_nearest_enemy(bot);
		if (enemy) {
			static const char *quips[] = {
				"Hi, %s!", "Wow, %s!", "Coommee heeereee, %s...",
				"Let's play soldiers, %s. :P", "%s: F33r m3, sux0r.",
				"Excellent, I can see a new training target!",
			};

			yell(bot, enemy, 5, quips, 6);
			bot->ai.state = AI_HUNT;
			bot->ai.hunt.pursuing = enemy;
			bot_hunt(bot);
			return;
		}
	}

	/* TODO: The exploring routine is quite trivial. We should chose one
	 * of the already known homes and sometimes head to it. We should
	 * also make use of already empty space. --pasky */

	if (--bot->ai.explore.tick_count < 0) {
		bot->ai.explore.tick_count = FPS * (mkrand(40) + 1) / 4;

		bot->tank->next_angle = mkrand(8);
		/* Try to make the struct ai less deadly - slow down when turning
		 * around. */
		speed /= delta_angle(bot->tank->next_angle, bot->tank->obj.angle) + 1;
	}

	/* Bhum dhum... */
	bot->tank->speed = speed;
}

void    bot_hunt (struct player *bot)
{
	struct player *victim = bot->ai.hunt.pursuing, *victim2;
	int angle;
	int speed = TANK_SPEED;

	assert(victim);

	/* Make sure there isn't a much better victim. */
	victim2 = find_nearest_enemy(bot);
	if (victim2
	    && distance(bot->tank->obj.x, bot->tank->obj.y, victim2->tank->obj.x, victim2->tank->obj.y) * RETARGET_RATIO
	       < distance(bot->tank->obj.x, bot->tank->obj.y, victim->tank->obj.x, victim->tank->obj.y)) {
		victim = bot->ai.hunt.pursuing = victim2;
	}

	/* Set the right angle if we didn't yet.
	 * Possibly change it a bit randomly. */
	angle = fuzzy_angle(get_angle(bot->tank->obj.x, bot->tank->obj.y,
	                              victim->tank->obj.x, victim->tank->obj.y),
	                    HUNT_FUZZINESS1, HUNT_FUZZINESS2);
	if (angle != bot->tank->obj.angle && bot->ai.last_turnaround > HUNT_TURN_PERIOD) {
		/* Try to make the struct ai less deadly - slow down when turning
		 * around. */
		speed /= delta_angle(angle, bot->tank->obj.angle) + 1;
		bot->tank->next_angle = angle;
		bot->ai.last_turnaround = 0;
	}

	/* Shooting angle? (We can sometimes hit the tank's corner even from
	 * a "wrong" angle.) */
	if (good_shot(bot->tank, victim->tank)) {
		bot->tank->fire = 1; /* Rat-ta-tat-ta-ra-ta-tat-ta-ta! */
	}

	/* Pursue on. */
	bot->tank->speed = speed;
}

void    bot_flee (struct player *bot)
{
	int angle;
	int speed = TANK_SPEED;

	AIDEBUG("[%d] Fleeing [%d,%d]->[%d,%d]...", bot->id,
	        bot->tank->obj.x, bot->tank->obj.y,
		bot->ai.flee.home_x, bot->ai.flee.home_y);

	/* When we are aligned with the gate, update y to point to the
	 * middle of the home. We assume the gate is at least 4 chars wide. */
	if (abs(bot->tank->obj.x - bot->ai.flee.home_x) <= 2*XPOINTS_PERCHAR) {
		bot->ai.flee.home_y = bot->home->obj.y;

		if (abs(bot->tank->obj.y - bot->ai.flee.home_y) <= 1*YPOINTS_PERCHAR) {
			bot->ai.state = AI_IDLE;
			bot->ai.idle.tick_count = FPS / 2;
			bot_idle(bot);
			return;
		}
	}

	/* TODO: Deal with tanks in the way (agressively >:). --pasky */

	/* Set the right angle if we didn't yet.
	 * No randomization nor speeding down, we must not waste time. */
	angle = get_angle(bot->tank->obj.x, bot->tank->obj.y,
	                  bot->ai.flee.home_x, bot->ai.flee.home_y);
	if (angle != bot->tank->obj.angle) {
		bot->tank->next_angle = angle;
	}

	/* If the energy is critical, use lower speed. */
	if (bot->tank->energy <= MAX_CRITICAL_ENERGY
	    || bot->tank->shields <= MAX_CRITICAL_SHIELDS)
		speed /= 2; /* maximal speed taking only 2PP per tick */

	/* Firing takes precious energy. */
	bot->tank->fire = 0;

	/* Aieeeeeeeeeee... */
	bot->tank->speed = speed;
}

/* Returns 0 if backing off was finished. */
int     bot_backoff (struct player *bot)
{
	int delta;

	/* Resume normal operation if we already got far enough. */
	delta = distance(bot->tank->obj.x, bot->tank->obj.y,
			 bot->ai.backoff_x, bot->ai.backoff_y);
	if (delta > bot->ai.backing_off * BACKOFF_DISTANCE * BACKOFF_DISTANCE) {
		AIDEBUG("[%d] backed off! d((%d,%d),(%d,%d))^2 = %d", bot->id,
		        bot->tank->obj.x, bot->tank->obj.y,
			bot->ai.backoff_x, bot->ai.backoff_y,
			delta);
		bot->ai.backing_off = 0;
		return 0;
	}

	AIDEBUG("[%d] backing off in angle %d...", bot->id, bot->tank->obj.angle);
	/* Angle is set in the stuck hook. If we fail in this
	 * direction, the stuck hook will strike again and turn
	 * around another time. Roar ahead! (Well, in fact, roar back. ;-) */
	bot->tank->speed = TANK_SPEED;
	bot->tank->fire = 0;
	return 1;
}

void    update_bot (struct tank *tank)
{
	struct player *bot = tank->player;
	static const state_hook states[] = {
		bot_idle,
		bot_explore,
		bot_hunt,
		bot_flee,
	};
	static const char *quips[] = {
		"La la la...", "Na na na.", "Pom pom.", "Brm brm...", "Grm!",
		"Im a little bottie, I will kill ya suckie!",
		"F33r m4 31337 h4x0r4 sk1ll2, 5ux0r2!", "M3 w1ll h4x0r y3r b4ttl3 d3v1c0r!",
		/* Explore mode only: */
		"So, where's anyone?!", "Come here, anyone!", "Time for lunch, I guess.",
	};

	if (!tank_ready(tank))
		return;

	if (mkrand(FPS * 10) == 0)
		yell(bot, NULL, 2, quips, 8 + (bot->ai.state == AI_EXPLORE ? 3 : 0));

	assert(bot->ai.state < AI_MAX);

	bot->ai.last_shot++;
	bot->ai.last_turnaround++;

	if (bot->ai.state != AI_FLEE && !surface_calc_collision(&tank->obj.spos, G_HOME)
	    && (tank->energy < (in_revenge(bot) ? MIN_REVENGE_ENERGY : MIN_NORMAL_ENERGY)
	        || tank->shields < (in_revenge(bot) ? MIN_REVENGE_SHIELDS : MIN_NORMAL_SHIELDS))) {
		/* Let's better go. */
		flee_home(bot);
	}

	if (!bot->ai.backing_off || !bot_backoff(bot))
		states[bot->ai.state](bot);
}
