#include <assert.h>
#include <stdlib.h>
#include "common/debug.h"
#include "common/globgame.h"
#include "common/math.h"
#include "common/object.h"
#include "common/surface.h"
#include "common/xmalloc.h"
#include "common/viewpoint.h"
#include "server/console.h"
#include "server/game.h"
#include "server/players.h"
#include "server/server.h"

struct tank tanks;
struct bullet bullets;
struct home homes;
struct fortress fortresses[MAX_TEAMS];
struct flag flags[MAX_TEAMS];

struct npc ferda;



static struct bullet *init_bullet (struct tank *tank);



/***************************************************************************/
/*                                   TANK                                  */
/***************************************************************************/

static struct tank *
init_tank(struct player *plr, int id)
{
	struct tank *tank = CALLOC(1, sizeof(struct tank));

	tank->status = TANK_READY;
	tank->score = tank->kills_in_line = 0;
	tank->deaths = 0;
	tank->player = plr;
	tank->color = id%8;

	add_object(&tank->obj);

	add_to_list(tanks, tank);
	return tank;
}

static void
done_tank(struct tank *tank)
{
	surface_drc_sprite(&tank->obj.spos, G_EMPTY, ~G_PLAYER);
	if (tank->status == TANK_FLAGGED)
		synth_flag(tank->obj.x, tank->obj.y, tank->flag_team);
	if (tank_solid(tank))
		del_object(&tank->obj);
	del_from_list(tank);
	FREE(tank);
}

static void
check_telefrag(struct tank *tank)
{
	struct spritepos cp = tank->obj.spos;
	struct player *q;

	foreach (q, players) {
		if (!q->tank || !tank_solid(q->tank) || tank->player == q)
			continue;
		if (!tank_in_region(q->tank, cp.sx, cp.sy, cp.sprite->width, cp.sprite->height))
			continue;
		bcast_telefrag(q, tank->player);
		kill_tank(q->tank, tank);
	}
}

void	spawn_tank(struct tank *tank, struct home *home)
{
	stop_ai_pursue(tank->player); /* Just to be sure... */
	if (player_is_ai(tank->player))
		reset_ai(&tank->player->ai);

	tank->status = TANK_READY;

	tank->obj.x = home->obj.x;
	tank->obj.y = home->obj.y;
	tank->next_angle = -1;
	tank->speed = tank->dirt_points = tank->dirt_penalty = 0;

	tank->fire = tank->firing = 0;
	tank->bullets_in_flight = tank->bullet_penalty = 0;

	tank->energy = MAX_ENERGY;
	tank->shields = MAX_SHIELDS;

	tank->obj.gfxid = GFX_TANK_ID;
	tank->obj.color = tank->color;
	tank->obj.angle = 0;
	obj_sync_sprite(&tank->obj);

	check_telefrag(tank);

	surface_drc_sprite(&(tank->obj.spos), G_PLAYER, ~G_DIRT);
}

void	explode_tank(struct tank *tank)
{
	stop_ai_pursue(tank->player);
	tank->status = TANK_EXPLODING;
	surface_drc_sprite(&tank->obj.spos, G_EMPTY, ~G_PLAYER);
	tank->obj.gfxid = GFX_EXPL_ID;
	tank->obj.angle = 0;
}

void    pause_tank(struct tank *tank)
{
	struct player *p = tank->player;
	struct spritepos *cp = &tank->obj.spos;
	struct spritepos *hp = &tank->home->obj.spos;

	DEBUG("Pausing player %d/%s, current state %d.", p->id, p->name, tank->status);

	assert(tank->status != TANK_PAUSED);
	if (!tank_ready(tank))
		return;

	/* Allow pausing only in own home and with shields and energy over 2/3. */
	if (!INTERSEC(cp->sx, cp->sy, cp->sprite->width, cp->sprite->height,
	              hp->sx, hp->sy, hp->sprite->width, hp->sprite->height)
	    || tank->energy < 2*MAX_ENERGY/3 || tank->shields < 2*MAX_SHIELDS/3)
		return;

	stop_ai_pursue(p); /* Move along. */
	tank->speed = 0;
	tank->status = TANK_PAUSED;
	tank->obj.color = 8;
	bcast_pause(p, 1);
}

void    unpause_tank(struct tank *tank)
{
	assert(tank->status == TANK_PAUSED);
	tank->status = TANK_READY;
	tank->obj.color = tank->color;
	bcast_pause(tank->player, 0);
}

void	hit_tank(struct tank *recv, struct tank *send, u16 points)
{
	DEBUG("TANK HITTED BY %d POINTS", points);

	if (!tank_ready(recv))
		return;

	recv->shields -= points;
	if (recv->shields <= 0) {
		bcast_kill(recv->player, send->player);
		kill_tank(recv, send);

	} else if (player_is_ai(recv->player)) {
		hit_ai_hook(recv->player, send->player);
	}
}

static void
update_tank (struct tank *tank)
{
	struct player *p = tank->player;
	struct spritepos new_pos, *cp = &tank->obj.spos;
	int got_stuck = 0;


	/* Explosion. */
	if (tank->status == TANK_EXPLODING) {
		if ((tick % (FPS/12)) == 0) {
			surface_drc_sprite(cp, G_EMPTY, ~G_DIRT);
			if (++tank->obj.angle > 7) {
				spawn_tank(tank, tank->home);
				if (tank->loser) tank->status = TANK_OBSERVER;
				return;
			}
			obj_sync_sprite(&tank->obj);
		}
		return;
	}


	/* Count collec^H^Hided dirt. */
	tank->dirt_points += surface_calc_collision(cp, G_DIRT);
	if (!player_is_ai(tank->player)) DEBUG("dirt_points %d", tank->dirt_points);

	/* Remove the player and the dirt. */
	surface_drc_sprite(cp, G_EMPTY, ~(G_DIRT | G_PLAYER));

	/* Take care of ghost tanks. */
	if (!tank_solid(tank)) {
		if (tank->obj.next) {
			del_object(&tank->obj);
			tank->obj.next = NULL;
		}
		return;

	} else {
		if (!tank->obj.next) {
			add_object(&tank->obj);
		}
	}


	/* Turn the tank. */
	if (tank->next_angle != -1) {
		rotsprite_by_obj(&tank->obj, &new_pos, tank->next_angle);
		if (! surface_calc_collision(&new_pos, G_STONE|G_PLAYER)) {
			tank->obj.angle = tank->next_angle;
			*cp = new_pos;
		} else {
			DEBUG("[%d] STUCK, need backoff", p->id);
			tank->speed = 0;
			got_stuck = 1;
		}
	}

	/* Calculate tank's speed and vector. */
	if (!(tick % DIRT_RECOVERY)) {
		tank->dirt_penalty = tank->dirt_points / DIRT_THRESHOLD;
		tank->dirt_points -= tank->dirt_points * DIRT_RECOVERY_RATIO;
	}
	get_vector(tank->obj.angle, tank->dirt_penalty ? tank->speed/(DIRT_SLOWING*tank->dirt_penalty)
	                                               : tank->speed, &tank->obj);

	/* Move the tank ahead. */
	nxtsprite_by_obj(&tank->obj, &new_pos);
	if (! surface_calc_collision(&new_pos, G_STONE|G_PLAYER)) {
		advance_object(&tank->obj);
		*cp = new_pos;
	} else {
		DEBUG("[%d] STUCKtoo, need backoff", p->id);
		tank->speed = 0;
		got_stuck = 1;
	}


	/* Project the player back. We project it _onto_ the potential dirt
	 * which we will accumulate and remove in the next frame. */
	surface_drc_sprite(cp, G_PLAYER, ~G_EMPTY);


	/* Fire. */
	if (tank->fire_burst < BULLET_BURST)
		tank->fire_burst++;
	if (tank->fire && tank->fire_burst >= BULLET_BURST && !tank->bullet_penalty) {
		if (tank->bullets_in_flight == (MAX_BULLETS-1))
			tank->bullet_penalty++;
		// tank->fire = 0;
		tank->fire_burst = 0;
		tank->energy -= FIRE_CONSUMPTION;
		DEBUG("new bullet");
		init_bullet(tank);
	}

	/* Adjust energy. */
	if (tank->speed > 0) {
		if (!(tick % MOVEMENT_CONSUMPTION))
			tank->energy -= 16 + tank->speed;
	} else if (!(tick % 8))
		tank->energy -= IDLE_CONSUMPTION;

	/* Hooks. */
	if (got_stuck && player_is_ai(p))
		stuck_ai_hook(p);
	if (tank->energy <= 0) {
		bcast_energydeath(tank->player);
		kill_tank(tank, NULL);
	}
}


/* Utility for bullets. */

int     tank_in_region(struct tank *tank, int x, int y, int wx, int wy)
{
	struct spritepos *cp = &tank->obj.spos;

	//DEBUG("x1(%d)y1(%d)w1(%d)h1(%d) x2(%d)y2(%d)w1(%d)h1(%d)", x,y,wx,wy,cp->sx,cp->sy
	return INTERSEC(x,y,wx,wy,cp->sx,cp->sy,cp->sprite->width,cp->sprite->height);
}

struct player	*find_player_by_xy (u16 x, u16 y, u16 wx, u16 wy)
{
	struct player	*p;

	foreach(p, players)
		/* Checking for p->tank to avoid races during connecting. */
		if (p->tank && tank_solid(p->tank) && tank_in_region(p->tank, x, y, wx, wy))
			return p;

	return NULL;
}



/***************************************************************************/
/*                                  BULLETA                                */
/***************************************************************************/

/* Teoreticky by se to dalo upravit na jakykoliv uhel, pokud bychom meli
 * dostupne sprity. */

static struct bullet *
init_bullet (struct tank *tank)
{
	struct bullet *b = CALLOC(1, sizeof(struct bullet));
	int sx, sy;

	b->obj.color = 0;
	b->obj.gfxid = GFX_BULLET_ID;

	b->obj.angle = tank->obj.angle;
	get_vector(b->obj.angle, tank->speed/2 + BULLET_SPEED, &b->obj);

	sprite_border(tank->obj.spos.sprite, tank->obj.angle, &sx, &sy);
	b->obj.x = tank->obj.x + sx*XPOINTS_PERCHAR;
	b->obj.y = tank->obj.y + sy*YPOINTS_PERCHAR;
	obj_sync_sprite(&b->obj);
	if (surface_calc_collision(&b->obj.spos, G_STONE)) {
		free(b);
		return NULL;
	}

	b->power = BULLET_POWER;
	b->owner = tank;
	tank->bullets_in_flight++;

	add_object(&b->obj);
	add_to_list(bullets, b);
	return b;
}

static void
done_bullet (struct bullet *b)
{
	del_object(&b->obj);
	del_from_list(b);
	FREE(b);
}

static void
update_bullet (struct bullet *b)
{
	struct spritepos new_pos = b->obj.spos, *cp = &b->obj.spos;
	u16 lower;

	surface_drc_sprite(cp, G_EMPTY, ~G_BULLET);

	nxtsprite_by_obj(&b->obj, &new_pos);

	if (surface_calc_collision(&new_pos, G_STONE)) {
		DEBUG("COLISION stone");
		goto delete;

	} else if (surface_calc_collision(&new_pos, G_PLAYER)) {
		struct player *colplr;

		DEBUG("COLISION player");
		colplr = find_player_by_xy(new_pos.sx, new_pos.sy, new_pos.sprite->width, new_pos.sprite->height);
		if (!colplr) {
			DEBUG("GAMERROR cannot find colliding player!");
			return;
		}
		if (!tank_solid(colplr->tank))
			return;
		b->refcount++;
		if (tank_ready(colplr->tank))
			hit_tank(colplr->tank, b->owner, b->power);
		b->refcount--;
		goto delete;
	}

	lower = surface_calc_collision(&new_pos, G_DIRT);
	b->power -= lower*DIRT_SLOWING;
	if (b->power < 0)
		goto delete;
	//DEBUG("lower: %d power: %d", lower, b->power);

	advance_object(&b->obj);
	*cp = new_pos;

	surface_drc_sprite(cp, G_BULLET, ~G_DIRT);
	return;

delete:
	if (!--b->owner->bullets_in_flight)
		b->owner->bullet_penalty = 0;

	/* TODO: Explode bullets. */
	done_bullet(b);
}



/***************************************************************************/
/*                                   HOME                                  */
/***************************************************************************/

static struct home *
init_home(struct player *plr)
{
	struct home *home = CALLOC(1, sizeof(struct home));
	int i;

	home->player = plr;
	home->obj.angle = 0;
	home->obj.gfxid = GFX_HOME_ID;
	home->obj.color = plr->tank->color;
	for (i = 0; i < PLACEMENT_MAXITERS; i++) {
		int team_id;
		int distdiv = teams_enforce - 1;

		while (1) {
			home->obj.x = mkrand((SURFACE_WIDTH-40)*XPOINTS_PERCHAR);
			home->obj.y = mkrand((SURFACE_HEIGHT-20)*YPOINTS_PERCHAR);

			obj_sync_sprite(&home->obj);

			if (!surface_calc_collision(&(home->obj.spos), G_STONE|G_HOME))
				break;
		}

		if (gstyle != GSTYLE_CTF)
			break;

		if (distdiv < 1) {
			distdiv = 0;
			for (team_id = 1; team_id < MAX_TEAMS; team_id++)
				if (teams[team_id].active)
					distdiv++;
			distdiv--;
		}
		if (distdiv < 1)
			distdiv = 1;

		for (team_id = 1; team_id < MAX_TEAMS; team_id++) {
			struct fortress *fort = &fortresses[team_id];
			int dist = distance(fort->obj.x, fort->obj.y,
			                    home->obj.x, home->obj.y);

			if (!teams[team_id].active)
				continue;

			if (dist < CTF_FORT_HOME_MINDIST * CTF_FORT_HOME_MINDIST / (distdiv * distdiv))
				goto next_iter;
		}

		break;
next_iter:;
	}
	if (i == PLACEMENT_MAXITERS)
		DEBUG("Home placement didn't succeed, random plcement after maxiters.");

	add_object(&home->obj);
	surface_drc_region(&home->obj.spos, G_HOME);
	/* We need to mask G_HOME out, otherwise client renders crap. */
	surface_drc_sprite(&home->obj.spos, G_STONE, ~(G_DIRT|G_HOME));

	add_to_list(homes, home);
	return home;
}

static void
done_home(struct home *home)
{
	del_object(&home->obj);
	surface_drc_region(&home->obj.spos, G_DIRT);
	del_from_list(home);
	FREE(home);
}

static void
update_home (struct home *home)
{
	struct player *q;
	struct spritepos *cp = &home->obj.spos, *np;

	if ((tick % (FPS/2)) != 0)
		return;

	foreach (q, players) {
		if (!q->tank) continue;
		np = &(q->tank->obj.spos);
		if (tank_ready(q->tank)
		    && INTERSEC(np->sx, np->sy, np->sprite->width-2, np->sprite->height-2,
		                cp->sx, cp->sy, cp->sprite->width-2, cp->sprite->height-2)) {
			u16 adde = (home->player == q ? ADDE_HOME : ADDE_FOREIGN);
			u16 adds = (home->player == q ? ADDS_HOME : 0);

			if (q->tank->energy < MAX_ENERGY && (q->tank->energy += adde) > MAX_ENERGY)
				q->tank->energy = MAX_ENERGY;
			if (q->tank->shields < MAX_SHIELDS && (q->tank->shields += adds) > MAX_SHIELDS)
				q->tank->shields = MAX_SHIELDS;
		}
	}
}



/***************************************************************************/
/*                                 FORTRESS                                */
/***************************************************************************/

static struct fortress *
init_fortress(int team_id)
{
	struct fortress *fort = &fortresses[team_id];
	int i;

	fort->team_id = team_id;

	fort->obj.gfxid = GFX_FORTRESS_ID;
	fort->obj.color = (7 + team_id) % 8;
	fort->obj.angle = team_id % gfx_objects[fort->obj.gfxid]->n;

	for (i = 0; i < PLACEMENT_MAXITERS; i++) {
		struct home *home;
		int team;
		int distdiv = teams_enforce - 1;

		while (1) {
			fort->obj.x = mkrand((SURFACE_WIDTH-40)*XPOINTS_PERCHAR);
			fort->obj.y = mkrand((SURFACE_HEIGHT-20)*YPOINTS_PERCHAR);

			obj_sync_sprite(&fort->obj);

			if (!surface_calc_collision(&(fort->obj.spos), G_STONE|G_HOME))
				break;
		}

		if (gstyle != GSTYLE_CTF)
			break;

		if (distdiv < 1) {
			distdiv = 0;
			for (team_id = 1; team_id < MAX_TEAMS; team_id++)
				if (teams[team_id].active)
					distdiv++;
			distdiv--;
		}
		if (distdiv < 1)
			distdiv = 1;

		for (team = 1; team < MAX_TEAMS; team++) {
			if (!teams[team].active || team == fort->team_id)
				continue;

			if (distance(fort->obj.x, fort->obj.y,
			             fortresses[team].obj.x,
			             fortresses[team].obj.y)
			    < (CTF_FORT_FORT_MINDIST * CTF_FORT_FORT_MINDIST)/(distdiv * distdiv))
				goto next_iter;
		}

		foreach (home, homes) {
			int dist = distance(fort->obj.x, fort->obj.y,
			                    home->obj.x, home->obj.y);

			if (dist < CTF_FORT_HOME_MINDIST * CTF_FORT_HOME_MINDIST / (distdiv * distdiv))
				goto next_iter;
		}

		break;
next_iter:;
	}
	if (i == PLACEMENT_MAXITERS)
		DEBUG("Fortress placement didn't succeed, random plcement after maxiters.");

	surface_drc_region(&fort->obj.spos, G_EMPTY);
	surface_drc_sprite(&fort->obj.spos, G_STONE, ~G_DIRT);
	add_object(&fort->obj);

	return fort;
}

#if 0
static void
done_fortress(struct fortress *fort)
{
	del_object(&fort->obj);
	surface_drc_region(&fort->obj.spos, G_DIRT);
}
#endif

static void
update_fortress (struct fortress *fort)
{
	struct player *q;
	struct spritepos *cp = &fort->obj.spos, *np;

	if ((tick % (FPS/2)) != 0)
		return;

	foreach (q, players) {
		if (!q->tank) continue;
		np = &(q->tank->obj.spos);
		if (tank_ready(q->tank)
		    && INTERSEC(np->sx, np->sy, np->sprite->width-2, np->sprite->height-2,
		                cp->sx, cp->sy, cp->sprite->width-2, cp->sprite->height-2)) {
			q->tank->energy  += (fort->team_id == q->team ? FADDE_HOME : FADDE_FOREIGN);
			q->tank->shields += (fort->team_id == q->team ? FADDS_HOME : 0);
			if (q->tank->energy > MAX_ENERGY)
				q->tank->energy = MAX_ENERGY;
			if (q->tank->shields > MAX_SHIELDS)
				q->tank->shields = MAX_SHIELDS;
		}
	}
}



/***************************************************************************/
/*                                   FLAG                                  */
/***************************************************************************/

struct flag *
synth_flag(int x, int y, int team_id)
{
	struct flag *flag = &flags[team_id];

	flag->team_id = team_id;

	flag->obj.gfxid = GFX_FLAG_ID;
	flag->obj.color = (7 + team_id) % 8;
	flag->obj.angle = team_id % gfx_objects[flag->obj.gfxid]->n;
	flag->obj.x = x;
	flag->obj.y = y;
	obj_sync_sprite(&flag->obj);
	add_object(&flag->obj);

	return flag;
}

static struct flag *
init_flag(int team_id)
{
	struct fortress *fort = &fortresses[team_id];

	return synth_flag(fort->obj.x, fort->obj.y, team_id);
}

static void
done_flag(struct flag *flag)
{
	del_object(&flag->obj);
	flag->team_id = -1; /* Mark invalid. */
}

static void
update_flag (struct flag *flag)
{
	struct tank *tank;
	struct spritepos *fpos = &flag->obj.spos;
	struct fortress *fort = &fortresses[flag->team_id];

	if (flag->team_id < 0)
		return; /* dead flag */

	foreach (tank, tanks) {
		struct spritepos *tpos = &tank->obj.spos;

		if (!INTERSEC(tpos->sx, tpos->sy, tpos->sprite->width-2, tpos->sprite->height-2,
		              fpos->sx, fpos->sy, fpos->sprite->width-2, fpos->sprite->height-2))
			continue;

		/* Collision! */

		if (tank->player->team != flag->team_id) {
			if (tank_ready(tank) && tank->status != TANK_FLAGGED) {
				/* Pick up the flag. */
				bcast_flagpick(tank->player, flag->team_id);

				tank->status = TANK_FLAGGED;
				tank->flag_team = flag->team_id;
				tank->obj.gfxid = GFX_FLAGTANK_ID;
				obj_sync_sprite(&tank->obj);

				teams[tank->flag_team].fstolen = 1;
				done_flag(flag);
				return;
			}
			continue;
		}

		/* Same team. */
		if (flag->obj.x != fort->obj.x || flag->obj.y != fort->obj.y) {
			/* Return the flag. */
			int team_id = flag->team_id;

			bcast_flagreturn(tank->player, flag->team_id);
			teams[flag->team_id].fstolen = 0;
			done_flag(flag);
			init_flag(team_id); /* flag->team_id invalid now */
			return;
		}

		/* It's at its place. */
		if (tank->status == TANK_FLAGGED && tank->player->team == flag->team_id) {
			/* Deliver the carried flag. */
			bcast_flagwin(tank->player, tank->flag_team);
			teams[flag->team_id].fwon++;
			teams[tank->flag_team].flost++;
			teams[tank->flag_team].fstolen = 0;
			init_flag(tank->flag_team);
			tank->flag_team = 0;
			tank->status = TANK_READY;
			tank->obj.gfxid = GFX_TANK_ID;
			obj_sync_sprite(&tank->obj);
			/* This did NOT affect the current flag,
			 * so we may just continue. */
		}
	}
}



/***************************************************************************/
/*                                NPC (FERDA)                              */
/***************************************************************************/

void    init_ferda(struct npc *npc)
{
	npc->speed = 4;
	npc->obj.angle = 1;
	npc->obj.gfxid = GFX_HOME_ID;
	npc->obj.color = 7;
	npc->dir = 0;

	while (1) {
		npc->obj.x = mkrand((SURFACE_WIDTH - 40) * XPOINTS_PERCHAR);
		npc->obj.y = mkrand((SURFACE_HEIGHT - 20) * YPOINTS_PERCHAR);
		obj_sync_sprite(&npc->obj);
		if (!surface_calc_collision(&(npc->obj.spos), G_STONE|G_HOME))
			break;
	}

	add_object(&npc->obj);
}

static void
runferda(void)
{
	struct npc	*npc = &ferda;
	u16 	i, best=8, min=100, max=0, eat;

	DEBUG("Ferda pos: %d %d\n", npc->obj.x, npc->obj.y);
	surface_drc_sprite(&npc->obj.spos, G_DIRT, ~G_STONE);
	for (i=0; i<8; i+=2) {
		struct spritepos new_pos;

		get_vector(i, 48 /* npc->speed, but wider radius */, &npc->obj);
		nxtsprite_by_obj(&npc->obj, &new_pos);

		if (surface_calc_collision(&new_pos, G_STONE|G_PLAYER|G_HOME))
			continue;

		eat = surface_calc_collision(&new_pos, G_DIRT);
		if (eat == new_pos.sprite->width * new_pos.sprite->height /* all dirt */)
			continue;

		if (eat < min) {
			min = eat;
			best = i;
		}
		if (eat > max) {
			max = eat;
		}
	}
	
	if (best != 8) {
		get_vector(best, npc->speed, &npc->obj);
		advance_object(&npc->obj);
		obj_sync_sprite(&npc->obj);
		surface_drc_sprite(&npc->obj.spos, G_STONE, ~G_DIRT);
	}
}



/***************************************************************************/
/*                                   PLAYER                                */
/***************************************************************************/

void	init_player(struct player *p, u32 id)
{
	p->tank = init_tank(p, id);
	p->home = p->tank->home = init_home(p);

	spawn_tank(p->tank, p->home);
}

void	finit_player(struct player *p)
{
	struct bullet *b, *n;

	foreachsafe(b, n, bullets) {
		if (!b->refcount && b->owner == p->tank)
			done_bullet(b);
	}
	stop_ai_pursue(p);

	done_home(p->home);
	p->home = p->tank->home = NULL;
	done_tank(p->tank);
	p->tank = NULL;
}

void	player_move (struct player *p, u8 direction, u8 speed, u8 firing)
{
	if (direction != p->tank->obj.angle)
		p->tank->next_angle = direction%8;
	p->tank->speed = speed == 0 ? 0 : 1 << (speed - 1);
	p->tank->fire = firing;
}



/***************************************************************************/
/*                                    GAME                                 */
/***************************************************************************/

void	init_game(void)
{
	init_list(tanks);
	init_list(bullets);
	init_list(homes);
	init_objects();
	if (do_ferda) init_ferda(&ferda);
}

void	init_ctf_team(int team_id)
{
	init_fortress(team_id);
	init_flag(team_id);
}

void    move_game(void)
{
	struct home *h, *hn;
	struct bullet *b, *bn;
	struct tank *tank, *tn;
	int i;

	foreachsafe(h, hn, homes)
		update_home(h);

	foreachsafe(b, bn, bullets)
		update_bullet(b);

	foreachsafe(tank, tn, tanks) {
		if (player_is_ai(tank->player))
			update_bot(tank);
		update_tank(tank);
	}

	if (gstyle == GSTYLE_CTF) {
		for (i = 1; i < MAX_TEAMS; i++) {
			if (!teams[i].active) continue;
			update_fortress(&fortresses[i]);
			update_flag(&flags[i]);
		}
	}

	if (do_ferda) runferda();
}



/****************************************************************************/

void	tank_view(struct tank *tank, struct viewpoint *v)
{
	calc_view(&tank->obj, v);
	v->en = tank->energy*255/MAX_ENERGY;
	v->sh = tank->shields*255/MAX_SHIELDS;
}



/****************************************************************************/

#if 0
void	dump_sprajt_list(void)
{
	struct spritepos *s;
	int i=0;
	foreach(s, sprites) {
		DEBUG("* sprajt sx: %d sy: %d id: %d ang: %d next:%x prev:%x", s->sx, s->sy, s->gfxid, s->angle, s->next, s->prev);
		i++;
	}
	DEBUG("total: %d", i);
}

void	dump_kulky_list(void)
{
	struct bullet *b;
	int i=0;
	DEBUG("me: %x next: %x prev %x", &bullets, bullets.next, bullets.prev);
	foreach(b, bullets) {
		DEBUG("* x: %d y: %d me: %x next: %x prev: %x", b->obj.x, b->obj.y, b, b->next, b->prev);
		i++;
	}
	DEBUG("total: %d", i);
}
#endif
