#include <assert.h>
#include <stdlib.h>
#include "common/bytef.h"
#include "common/debug.h"
#include "common/globgame.h"
#include "common/math.h"
#include "common/object.h"
#include "common/packet.h"
#include "common/xmalloc.h"
#include "server/console.h"
#include "server/game.h"
#include "server/players.h"
#include "server/server.h"


struct player 	players;
int	num_players;

struct team teams[MAX_TEAMS];

struct dyn_array bot_names = { NULL, 0 };


static void
kill_in_line(struct tank *t)
{
	t->kills_in_line++; DEBUG("%d has %d kills in line now for his glamor.", t->player->id, t->kills_in_line);
	if (t->kills_in_line == 4) bcast_glamorize(t->player, 1);
	else if (t->kills_in_line == 8) bcast_glamorize(t->player, 2);
	else if (t->kills_in_line == 16) bcast_glamorize(t->player, 3);
	else if (t->kills_in_line == 32) bcast_glamorize(t->player, 4); /* If someone attains this... */
}

static void
reset_kills_in_line(struct tank *t)
{
	if (t->kills_in_line > 3) bcast_glamorize(t->player, 0);
	t->kills_in_line = 0;
}

void    kill_tank(struct tank *t, struct tank *by)
{
	if (t->status == TANK_FLAGGED) {
		bcast_flagdrop(t->player, t->flag_team);
		synth_flag(t->obj.x, t->obj.y, t->flag_team);
		t->status = TANK_READY;
	}

	/* Keep in sync with client code! */
	if (!by || by == t || (t->player->team && t->player->team == by->player->team)) {
		if (!by) by = t;
		if (by->score >= 0)
			by->score--;

		if (teams[by->player->team].score >= 0)
			teams[by->player->team].score--;
	} else {
		kill_in_line(by);
		by->score++;

		teams[by->player->team].score++;
	}

	t->deaths++;
	teams[t->player->team].deaths++;
	reset_kills_in_line(t);
	explode_tank(t);

	if (by && by != t && player_is_ai(by->player))
		kill_ai_hook(by->player, t->player);
	if (player_is_ai(t->player))
		death_ai_hook(t->player, by ? by->player : NULL);

	if (gstyle == GSTYLE_LMS) {
		assert(t->deaths <= lms_lives);
		if (t->deaths == lms_lives) {
			/* It's over for you! */
			CONSOLE("%d::%s is out of the game!", t->player->id, t->player->name);
			t->loser = 1;
			check_winner();
		}
	}
}


struct player	*alloc_client (struct conn *tcp)
{
	struct player *p;
	p = (struct player *) CALLOC(1, sizeof(struct player));
	p->id = newobj();
	p->initialized = 0;
	p->tcp = tcp;

	add_to_list(players, p);
	return p;
}

struct player   *alloc_bot (void)
{
	struct player *p;
	p = (struct player *) CALLOC(1, sizeof(struct player));
	p->id = newobj();
	p->initialized = 0;
	reset_ai(&p->ai);

	add_to_list(players, p);
	return p;
}

void    join_team(struct player *p, int teamid)
{
	if (teamid >= MAX_TEAMS)
		return;

	if (teams_enforce > 0 && teamid == 0) {
		/* teamid 0 not allowed; get the smallest team. */
		int minplayers = INT_MAX, minteam = 0;

		assert(teams_enforce < MAX_TEAMS);
		for (teamid = 1; teamid <= teams_enforce; teamid++) {
			if (teams[teamid].players < minplayers) {
				minplayers = teams[teamid].players;
				minteam = teamid;
			}
		}
		assert(minteam > 0);
		teamid = minteam;
		/* Otherwise we just said it in bcast_join(). */
		bcast_teamup(p, teamid);
	}

	if (p->team != 0) teams[p->team].players--;
	if (teamid != 0) teams[teamid].players++;
	p->team = teamid;

	if (teamid) {
		if (gstyle == GSTYLE_CTF && !teams[teamid].active)
			init_ctf_team(teamid);
		teams[teamid].active++;
		p->tank->color = (7+teamid)%8;
	} else {
		p->tank->color = p->id%8;
	}
	p->tank->obj.color = p->tank->color;
	p->tank->home->obj.color = p->tank->color;
}

void	join (struct player *p)
{
	char	buffer[MAX_MSGLEN];
	int     blen = 0;

	if (p->initialized)
		return;
	p->cur_seq = p->last_seq = 0;
	p->initialized = 1;

	init_player(p, p->id);
	init_q(&(p->state), p->tank->obj.spos.sx, p->tank->obj.spos.sy);
	num_players++;

	if (!player_is_ai(p)) {
		switch (gstyle) {
			case GSTYLE_DM:
			case GSTYLE_CTF:
				bprintf(buffer, 4, "ww", CH_MSGGSTYLE, gstyle);
				blen = 4;
				break;
			case GSTYLE_LMS:
				bprintf(buffer, 6, "www", CH_MSGGSTYLE, gstyle, lms_lives);
				blen = 6;
				break;
			default: assert(0); break;
		}
		unicast(p, buffer, blen);
	}

	bcast_join(p);
}

void	kick (struct player *p)
{
	if (p->initialized) {
		bcast_leave(p);
		if (p->team != 0) teams[p->team].players--;
		num_players--;
		rmall_q(&(p->state));
		finit_player(p);
		free(p->name);

		if (p->udp) {
			p->udp->fd = -1;
			udp_destroy(p->udp);
		}
	}
	if (p->tcp) conn_destroy(p->tcp);
	del_from_list(p);
	FREE(p);

	check_winner();
}

static char *
name_bot (void)
{
	static int avail_names = INT_MAX;
	char dyn[64];

	if (avail_names == INT_MAX)
		avail_names = bot_names.len;

	if (avail_names <= 0) {
		snprintf(dyn, 16, "[b] %c%d", 'A' + (char)mkrand('Z' - 'A' + 1), -avail_names);
		avail_names--;
	} else {
		int i;

		do { i = mkrand(bot_names.len); } while (bot_names.arr[i] == NULL);
		snprintf(dyn, 64, "[b] %s", (char *) bot_names.arr[i]);
		avail_names--; bot_names.arr[i] = NULL;
	}
	return strdup(dyn);
}

void    spawn_bot (void)
{
	struct player *bot = alloc_bot ();
	int xteams_enforce = 0;

	/* Figure out some name. */
	bot->name = name_bot ();
	join(bot);

	if (teams_ai) { /* XXX */
		xteams_enforce = teams_enforce;
		teams_enforce = teams_ai;
	}
	join_team(bot, 0);
	if (xteams_enforce)
		teams_enforce = xteams_enforce;
}

void    check_winner(void)
{
	struct player *plr, *winner = NULL;

	if (gstyle != GSTYLE_LMS)
		return; /* Noone wins. */

	foreach (plr, players) {
		//DEBUG("id %d s %d", plr->id, plr->tank->status);
		if (!tank_ready(plr->tank))
			continue;
		assert(!plr->tank->loser);
		if (winner) {
			return;
		} else {
			winner = plr;
		}
	}
	if (!winner) {
		/* No players. */
		return;
	}

	bcast_winner(winner);

	DEBUG("NEXT ROUND");
	/* Another round. */
	foreach (plr, players) {
		plr->tank->loser = 0;
		plr->tank->score = plr->tank->deaths = 0;
		plr->tank->kills_in_line = 0;
		if (plr->tank->status != TANK_EXPLODING)
			plr->tank->status = TANK_READY;
		else
			plr->tank->deaths = -1; /* will be +1'd */
	}
}

void    init_players(void)
{
	init_list(players);
}

void    init_teams(void)
{
	int i;

	for (i = 0; i < MAX_TEAMS; i++) {
		char buf[MAX_NICKLEN];

		snprintf(buf, MAX_NICKLEN, "Team %d", i);
		memset(&teams[i], 0, sizeof(struct team));
		teams[i].name = strdup(buf);
	}
}
